import argh
import os
import json
import re
import subprocess
import tempfile
import sys
import shutil
import contextlib
import logging
import random
import time
import gzip
import io
import tarfile
from typing import Dict, Any
from urllib.request import urlopen
from .exceptions import NoHttpAddressAvailable


log = logging.getLogger(__file__)


RELEASE_URL = 'https://cdn.crate.io/downloads/releases/crate-{version}.tar.gz'
VERSION_RE = re.compile('^(\d+\.\d+\.\d+)$')
DEFAULT_SETTINGS = {
    'cluster.routing.allocation.disk.watermark.low': '1b',
    'cluster.routing.allocation.disk.watermark.high': '1b',
    'discovery.initial_state_timeout': 0,
    'discovery.zen.ping.multicast.enabled': False,
    'network.host': '127.0.0.1',
    'udc.enabled': False
}

HTTP_ADDRESS_RE = re.compile(
    '.*\[http +\] \[.*\] .*'
    'publish_address {'
    '(?:inet\[[\w\d\.-]*/|\[)?'
    '(?:[\w\d\.-]+/)?'
    '(?P<addr>[\d\.:]+)'
    '(?:\])?'
    '}'
)


class Timeout:
    def __init__(self, timeout, sleep=0.1):
        self.start_time = time.time()
        self.sleep = sleep
        self._first_ok = True
        self.timeout = timeout

        def timeout_expired():
            if self._first_ok:
                self._first_ok = False
                return False
            now = time.time()
            if (now - self.start_time) > self.timeout:
                return True
            if self.sleep:
                time.sleep(self.sleep)

        self._timeout_expired = timeout_expired

    def __call__(self):
        return not self._timeout_expired()


def wait_until(f, condition):
    """Calls f while condition returns True until f returns a truthy value

    >>> wait_until(lambda: True, lambda: True)
    """
    while condition():
        r = f()
        if r:
            break


def _get_http_uri(lines):
    """Get the http publish_address from lines that contain Crate log output.

    >>> lines = [
    ...     "[2016-06-11 19:10:16,141][INFO ][node                     ] [Ankhi] initialized",
    ...     "[2016-06-11 19:10:16,141][INFO ][node                     ] [Ankhi] starting ...",
    ...     "[2016-06-11 19:10:16,171][INFO ][http                     ] [Ankhi] publish_address {10.68.2.10:4200}, bound_addresses {[::]:4200}",
    ...     "[2016-06-11 19:10:16,188][INFO ][discovery                ] [Ankhi] crate/GJAvonoFSfS3Y1IaUPTqfA"
    ... ]
    >>> _get_http_uri(lines)
    'http://10.68.2.10:4200'

    >>> lines = [
    ...     "[2016-06-11 21:26:53,798][INFO ][node                     ] [Rex Mundi] starting ...",
    ...     "[2016-06-11 21:26:53,828][INFO ][http                     ] [Rex Mundi] bound_address {inet[/0:0:0:0:0:0:0:0:4200]}, publish_address {inet[/192.168.0.19:4200]}",
    ... ]
    >>> _get_http_uri(lines)
    'http://192.168.0.19:4200'

    >>> lines = [
    ...     "[2016-06-15 22:18:36,639][INFO ][node                     ] [crate] starting ...",
    ...     "[2016-06-15 22:18:36,755][INFO ][http                     ] [crate] publish_address {localhost/127.0.0.1:42203}, bound_addresses {127.0.0.1:42203}, {[::1]:42203}, {[fe80::1]:42203}",
    ...     "[2016-06-15 22:18:36,774][INFO ][transport                ] [crate] publish_address {localhost/127.0.0.1:4300}, bound_addresses {127.0.0.1:4300}, {[::1]:4300}, {[fe80::1]:4300}",
    ...     "[2016-06-15 22:18:36,779][INFO ][discovery                ] [crate] Testing42203/Eroq_ZAgT4CDpF_gzh4tcA",
    ... ]
    >>> _get_http_uri(lines)
    'http://127.0.0.1:42203'

    >>> lines = [
    ...     "[2016-06-16 10:27:20,074][INFO ][node                     ] [Selene] starting ...",
    ...     "[2016-06-16 10:27:20,150][INFO ][http                     ] [Selene] bound_address {inet[/192.168.43.105:4200]}, publish_address {inet[Haudis-MacBook-Pro.local/192.168.43.105:4200]}",
    ...     "[2016-06-16 10:27:20,165][INFO ][transport                ] [Selene] bound_address {inet[/192.168.43.105:4300]}, publish_address {inet[Haudis-MacBook-Pro.local/192.168.43.105:4300]}",
    ...     "[2016-06-16 10:27:20,185][INFO ][discovery                ] [Selene] crate/h9moKMrATmCElYXjfad5Vw",
    ... ]
    >>> _get_http_uri(lines)
    'http://192.168.43.105:4200'
    """
    for line in lines:
        m = HTTP_ADDRESS_RE.match(line)
        if m:
            return 'http://' + m.group('addr')


def _wait_until_reachable(url):
    def cluster_ready():
        try:
            with urlopen(url) as r:
                p = json.loads(r.read().decode('utf-8'))
                return int(p['status']) == 200
        except Exception as e:
            return False
    wait_until(cluster_ready, Timeout(60))


def _get_settings(settings=None):
    s = DEFAULT_SETTINGS.copy()
    if settings:
        s.update(settings)
    return s


class CrateNode(contextlib.ExitStack):
    """Class that allows starting and stopping a Crate process

    This is similar to the ``CrateLayer`` in ``crate.testing.layer``.
    But additionaly it supports setting environment variables and it can infer
    the port to which Crate binds by sniffing Crate's stdout.

    Attributes:
        http_url: The HTTP URL of the Crate process.
            Only available after ``start()`` has been called.
        process: The subprocess. Only available after ``start()`` has been called.
    """

    def __init__(self,
                 crate_dir: str,
                 data_path: str=None,
                 env: Dict[str, Any]=None,
                 settings: Dict[str, Any]=None) -> None:
        """Create a CrateNode

        Args:
            crate_dir: Path to the extracted Crate tarball
            env: Environment variables with which the Crate process will be
                started.
            settings: Additional Crate settings.
        """
        super().__init__()
        self.crate_dir = crate_dir
        self.env = env
        self.process = None  # type: subprocess.Popen
        self.http_url = None  # type: str
        start_script = 'crate.bat' if sys.platform == 'win32' else 'crate'

        settings = _get_settings(settings)
        self.data_path = settings.get('path.data') or tempfile.mkdtemp()
        self.logs_path = settings.get('path.logs') or os.path.join(crate_dir, 'logs')
        self.cluster_name = settings.get('cluster.name') or 'cr8'
        log.info('Work dir: %s (removed on stop)', self.data_path)
        settings['path.data'] = self.data_path
        settings['cluster.name'] = self.cluster_name
        args = ['-Des.{0}={1}'.format(k, v) for k, v in settings.items()]
        self.cmd = [
            os.path.join(crate_dir, 'bin', start_script)] + args

    def start(self):
        """Start the process.

        This will block until the Crate cluster is ready to process requests.
        """
        log.info('Starting Crate process')
        self.process = proc = self.enter_context(subprocess.Popen(
            self.cmd,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            env=self.env,
            universal_newlines=True
        ))
        log.info('PID: %s', proc.pid)
        self._obtain_http_url()
        log.info('HTTP: %s', self.http_url)
        _wait_until_reachable(self.http_url)
        log.info('Cluster is ready')

    def stop(self):
        if self.process:
            self.process.terminate()
            self.process.communicate(timeout=10)
        shutil.rmtree(self.data_path)

    def __enter__(self):
        return self

    def __exit__(self, *ex):
        self.stop()

    def _obtain_http_url(self):
        """
        Read Crate log file to obtain HTTP address of node
        """
        def wait(waited, wait_time=0.25):
            time.sleep(wait_time)
            return waited + wait_time

        logfile = os.path.join(self.logs_path, self.cluster_name + '.log')

        def logfile_exists():
            return os.path.exists(logfile)
        wait_until(logfile_exists, Timeout(10))

        time_waited = 0
        with open(logfile, encoding='utf-8') as fp:
            pos = 0
            while not self.http_url:
                fp.seek(pos)
                if pos > 10000 or time_waited > 30:
                    # don't wait forever
                    # if HTTP url could not be obtained within the first 10000 bytes
                    # or within 30 seconds it's probably not there
                    raise NoHttpAddressAvailable("Couldn't get HTTP URL")
                self.http_url = _get_http_uri(iter(fp.readline, ''))
                pos = fp.tell()
                time_waited = wait(time_waited)


def _download_and_extract(uri, crate_root):
    filename = os.path.basename(uri)
    crate_folder_name = re.sub('\.tar(\.gz)?$', '', filename)
    crate_dir = os.path.join(crate_root, crate_folder_name)
    if os.path.exists(crate_dir):
        log.info('Skipping download, tarball alrady extracted at %s', crate_dir)
        return crate_dir
    log.info('Downloading %s and extracting to %s', uri, crate_root)
    with io.BytesIO(urlopen(uri).read()) as tmpfile:
        with tarfile.open(fileobj=tmpfile) as t:
            t.extractall(crate_root)
    return crate_dir


def init_logging():
    log.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    log.addHandler(ch)


def _from_versions_json(key):
    def retrieve():
        with urlopen('https://crate.io/versions.json') as r:
            if r.headers.get('Content-Encoding') == 'gzip':
                with gzip.open(r, 'rt') as r:
                    versions = json.loads(r.read())
            else:
                versions = json.loads(r.read().decode('utf-8'))
        return versions[key]
    return retrieve


NIGHTLY_RE = re.compile('.*>(?P<filename>crate-\d+\.\d+\.\d+-\d{12}-[a-z0-9]{7,}\.tar\.gz)<.*')


def _find_last_nightly(lines):
    """Return the last nightly release tarball filename.

    >>> lines = [
    ...     '<a href="crate-0.55.0-201606080301-3ceb1ed.tar.gz">crate-0.55.0-201606080301-3ceb1ed.tar.gz</a>           08-Jun-2016 01:01            46298304'
    ...     '<a href="crate-0.55.0-201606090301-b32a36f.tar.gz">crate-0.55.0-201606090301-b32a36f.tar.gz</a>           09-Jun-2016 01:01            46297737'
    ...     '<a href="crate-0.55.0-201606100301-23388dd.tar.gz">crate-0.55.0-201606100301-23388dd.tar.gz</a>           10-Jun-2016 01:01            46300496'
    ... ]
    >>> _find_last_nightly(lines)
    'crate-0.55.0-201606100301-23388dd.tar.gz'
    """
    for line in reversed(lines):
        m = NIGHTLY_RE.match(line)
        if m:
            return m.group('filename')
    raise ValueError("Couldn't find a valid nightly tarball filename in the lines")


def _get_latest_nightly_uri():
    base_uri = 'https://cdn.crate.io/downloads/releases/nightly/'
    with urlopen(base_uri) as r:
        filename = _find_last_nightly([line.decode('utf-8') for line in r])
        return base_uri + filename


_version_lookups = {
    'latest': _from_versions_json('crate-java'),
    'latest-stable': _from_versions_json('crate-java'),
    'latest-testing': _from_versions_json('crate_testing'),
    'latest-nightly': _get_latest_nightly_uri
}


def _lookup_uri(version):
    if version in _version_lookups:
        version = _version_lookups[version]()
    m = VERSION_RE.match(version)
    if m:
        return RELEASE_URL.format(version=m.group(0))
    else:
        return version


def get_crate(version, crate_root=None):
    """Retrieve a Crate tarball, extract it and return the path.

    Args:
        version: The Crate version to get.
            Can be specified in different ways:

            - A concrete version like '0.55.0'
            - An alias: 'latest-stable' or 'latest-testing'
            - A URI pointing to a crate tarball
        crate_root: Where to extract the tarball to.
            If this isn't specified ``$XDG_CACHE_HOME/.cache/cr8/crates``
            will be used.
    """
    uri = _lookup_uri(version)
    crate_root = crate_root or os.environ.get(
        'XDG_CACHE_HOME', os.path.join(os.path.expanduser('~'), '.cache', 'cr8', 'crates'))
    crate_dir = _download_and_extract(uri, crate_root)
    return crate_dir


@argh.arg('version', help='Crate version to run. Concrete version like\
          "0.55.0", an alias or an URI pointing to a Crate tarball. Supported\
          aliases are: [{0}]'.format(', '.join(_version_lookups.keys())))
@argh.arg('-e', '--env', action='append')
@argh.arg('-s', '--setting', action='append')
def run_crate(version, env=None, setting=None, crate_root=None):
    """Launch a crate instance"""
    init_logging()
    settings = {
        'cluster.name': 'cr8-crate-run' + str(random.randrange(1e9))
    }
    crate_dir = get_crate(version, crate_root)
    if setting:
        settings.update(dict(i.split('=') for i in setting))
    if env:
        env = dict(i.split('=') for i in env)
    with CrateNode(crate_dir=crate_dir, env=env, settings=settings) as node:
        try:
            node.start()
            node.process.communicate()
        except NoHttpAddressAvailable as ex:
            print(ex)
        except KeyboardInterrupt:
            print('Stopping Crate...')


if __name__ == "__main__":
    argh.dispatch_command(run_crate)



# vi: set fileencoding=utf-8
# -*- coding: utf-8; -*-


class NoHttpAddressAvailable(Exception):
    """
    Exception is raised when no HTTP address could be obtained from logfile
    on node start within given time or log size.
    """
    pass




import argh
import os
import itertools
from time import time
from functools import partial
from crate.client import connect

from cr8 import aio
from .insert_json import to_insert
from .bench_spec import load_spec
from .timeit import QueryRunner, Result
from .misc import as_bulk_queries, as_statements, get_lines
from .metrics import Stats
from .cli import dicts_from_lines
from . import clients


BENCHMARK_TABLE = '''
create table if not exists benchmarks (
    version_info object (strict) as (
        number string,
        hash string,
        date timestamp
    ),
    statement string,
    meta object as (
        name string
    ),
    started timestamp,
    ended timestamp,
    concurrency int,
    bulk_size int,
    runtime_stats object (strict) as (
        avg double,
        min double,
        max double,
        mean double,
        median double,
        percentile object as (
            "50" double,
            "75" double,
            "90" double,
            "99" double,
            "99_9" double
        ),
        n integer,
        variance double,
        stdev double,
        samples array(double)
    )
) clustered into 8 shards with (number_of_replicas = '1-3', column_policy='strict')
'''


def _parse_version(version: str) -> tuple:
    if not version:
        return None
    major, minor, patch = version.split('.', maxsplit=3)
    return (int(major), int(minor), int(patch))


class Executor:
    def __init__(self, spec_dir, benchmark_hosts, result_hosts, output_fmt=None):
        self.benchmark_hosts = benchmark_hosts
        self.spec_dir = spec_dir
        self.conn = connect(benchmark_hosts)
        self.client = clients.client(benchmark_hosts)
        self.output_fmt = output_fmt
        self.server_version_info = aio.run(self.client.get_server_version)
        self.server_version = _parse_version(self.server_version_info['number'])

        if result_hosts:
            table_created = []

            def process_result(result):
                with connect(result_hosts) as conn:
                    cursor = conn.cursor()
                    if not table_created:
                        cursor.execute(BENCHMARK_TABLE)
                        table_created.append(None)
                    stmt, args = to_insert('benchmarks', result.as_dict())
                    cursor.execute(stmt, args)
                print(result)
                print('')
        else:
            def process_result(result):
                print(result)
                print('')
        self.process_result = process_result

    def _to_inserts(self, data_spec):
        target = data_spec['target']
        source = os.path.join(self.spec_dir, data_spec['source'])
        dicts = dicts_from_lines(get_lines(source))
        return (to_insert(target, d) for d in dicts)

    def exec_instructions(self, instructions):
        cursor = self.conn.cursor()
        filenames = instructions.statement_files
        filenames = (os.path.join(self.spec_dir, i) for i in filenames)
        lines = (line for fn in filenames for line in get_lines(fn))
        statements = itertools.chain(as_statements(lines), instructions.statements)
        for stmt in statements:
            cursor.execute(stmt)

        for data_file in instructions.data_files:
            inserts = as_bulk_queries(self._to_inserts(data_file),
                                      data_file.get('bulk_size', 5000))
            concurrency = data_file.get('concurrency', 25)
            aio.run_many(self.client.execute_many, inserts, concurrency=concurrency)
            cursor.execute('refresh table {target}'.format(target=data_file['target']))

    def run_load_data(self, data_spec, meta=None):
        inserts = self._to_inserts(data_spec)
        statement = next(iter(inserts))[0]
        bulk_size = data_spec.get('bulk_size', 5000)
        inserts = as_bulk_queries(self._to_inserts(data_spec), bulk_size)
        concurrency = data_spec.get('concurrency', 25)
        num_records = data_spec.get('num_records', None)
        if num_records:
            num_records = max(1, int(num_records / bulk_size))
        stats = Stats()
        f = partial(aio.measure, stats, self.client.execute_many)
        start = time()
        aio.run_many(f,
                     inserts,
                     concurrency=concurrency,
                     num_items=num_records)
        end = time()
        self.process_result(Result(
            version_info=self.server_version_info,
            statement=statement,
            meta=meta,
            started=start,
            ended=end,
            stats=stats,
            concurrency=concurrency,
            bulk_size=bulk_size,
            output_fmt=self.output_fmt
        ))

    def _skip_message(self, min_version, stmt):
        msg = ('## Skipping (Version {server_version} instead of {min_version}):\n'
               '   Statement: {statement:.70}')
        msg = msg.format(
            statement=stmt,
            min_version='.'.join((str(x) for x in min_version)),
            server_version='.'.join((str(x) for x in self.server_version)))
        return msg

    def run_queries(self, queries, meta=None):
        for query in queries:
            stmt = query['statement']
            iterations = query.get('iterations', 1)
            concurrency = query.get('concurrency', 1)
            min_version = _parse_version(query.get('min_version'))
            if min_version and min_version > self.server_version:
                print(self._skip_message(min_version, stmt))
                continue
            print(('\n## Running Query:\n'
                   '   Statement: {statement:.70}\n'
                   '   Concurrency: {concurrency}\n'
                   '   Iterations: {iterations}'.format(
                       statement=str(stmt),
                       iterations=iterations,
                       concurrency=concurrency)))
            with QueryRunner(
                stmt,
                meta,
                repeats=iterations,
                hosts=self.benchmark_hosts,
                concurrency=concurrency,
                args=query.get('args'),
                bulk_args=query.get('bulk_args'),
                output_fmt=self.output_fmt
            ) as runner:
                result = runner.run()
            self.process_result(result)

    def __enter__(self):
        return self

    def __exit__(self, *ex):
        self.conn.close()
        self.client.close()


@argh.arg('benchmark_hosts', type=str)
@argh.arg('-of', '--output-fmt', choices=['full', 'short'], default='full')
@argh.wrap_errors([KeyboardInterrupt])
def run_spec(spec, benchmark_hosts, result_hosts=None, output_fmt=None):
    """Run a spec file, executing the statements on the benchmark_hosts.

    Short example of a spec file:

        [setup]
        statement_files = ["sql/create_table.sql"]

            [[setup.data_files]]
            target = "t"
            source = "data/t.json"

        [[queries]]
        statement = "select count(*) from t"
        iterations = 2000
        concurrency = 10

        [teardown]
        statements = ["drop table t"]

    See https://github.com/mfussenegger/cr8/tree/master/specs
    for more examples.

    Args:
        spec: path to a spec file
        benchmark_hosts: hostname[:port] pairs of Crate nodes
        result_hosts: optional hostname[:port] Crate node pairs into which the
            runtime statistics should be inserted.
        output_fmt: output format
    """
    with Executor(
        spec_dir=os.path.dirname(spec),
        benchmark_hosts=benchmark_hosts,
        result_hosts=result_hosts,
        output_fmt=output_fmt
    ) as executor:
        spec = load_spec(spec)
        try:
            print('# Running setUp')
            executor.exec_instructions(spec.setup)
            print('# Running benchmark')
            if spec.load_data:
                for data_spec in spec.load_data:
                    executor.run_load_data(data_spec, spec.meta)
            else:
                executor.run_queries(spec.queries, spec.meta)
        finally:
            print('# Running tearDown')
            executor.exec_instructions(spec.teardown)


def main():
    argh.dispatch_command(run_spec)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import json
import argh
import argparse
import operator
from faker import Factory
from functools import partial
from collections import OrderedDict
from crate.client import connect
from concurrent.futures import ProcessPoolExecutor

from .insert_json import to_insert
from .misc import parse_table
from .aio import asyncio, consume
from .cli import to_int
from .fake_providers import GeoSpatialProvider, auto_inc
from . import clients

loop = asyncio.get_event_loop()


def retrieve_columns(cursor, schema, table):
    cursor.execute(
        'select column_name, data_type from information_schema.columns \
        where is_generated = false and schema_name = ? and table_name = ? \
        order by ordinal_position asc', (schema, table))
    return OrderedDict(cursor.fetchall())


def generate_row(fakers):
    return [x() for x in fakers]


def x1000(func):
    return func() * 1000


def timestamp(fake):
    # return lamda: fake.unix_time() * 1000 workaround:
    # can't use lambda or nested functions because of multiprocessing pickling
    return partial(x1000, fake.unix_time)


class DataFaker:
    _mapping = {
        ('id', 'string'): operator.attrgetter('uuid4'),
        ('id', 'integer'): auto_inc,
        ('id', 'long'): auto_inc,
    }

    _type_default = {
        'integer': operator.attrgetter('random_int'),
        'long': operator.attrgetter('random_int'),
        'float': operator.attrgetter('pyfloat'),
        'double': operator.attrgetter('pydecimal'),
        'ip': operator.attrgetter('ipv4'),
        'timestamp': timestamp,
        'string': operator.attrgetter('word'),
        'boolean': operator.attrgetter('boolean'),
        'geo_point': operator.attrgetter('geo_point'),
    }

    _custom = {
        'auto_inc': auto_inc
    }

    def __init__(self):
        self.fake = Factory.create()
        self.fake.add_provider(GeoSpatialProvider)

    def provider_for_column(self, column_name, data_type):
        provider = getattr(self.fake, column_name, None)
        if provider:
            return provider
        custom_provider = self._custom.get(column_name)
        if custom_provider:
            return custom_provider(self.fake)
        alternative = self._mapping.get((column_name, data_type))
        if not alternative:
            alternative = self._type_default.get(data_type)
            if not alternative:
                msg = 'No fake provider found for column "{col}" with type "{type}"'
                raise ValueError(msg.format(col=column_name, type=data_type))
        return alternative(self.fake)

    def provider_from_mapping(self, column_name, mapping):
        key = mapping[column_name]
        args = None
        if isinstance(key, list):
            key, args = key
        provider = getattr(self.fake, key, None)
        if not provider:
            raise KeyError('No fake provider with name "%s" found' % (key,))
        if args:
            provider = partial(provider, *args)
        return provider


def create_row_generator(columns, mapping=None):
    fake = DataFaker()
    fakers = []
    for column_name, type_name in columns.items():
        if mapping and column_name in mapping:
            fakers.append(fake.provider_from_mapping(column_name, mapping))
        else:
            fakers.append(fake.provider_for_column(column_name, type_name))
    return partial(generate_row, fakers)


def generate_bulk_args(generate_row, bulk_size):
    return [generate_row() for i in range(bulk_size)]


async def _produce_data_and_insert(q, client, stmt, bulk_args_fun, num_inserts):
    executor = ProcessPoolExecutor()
    for i in range(num_inserts):
        args = await asyncio.ensure_future(
            loop.run_in_executor(executor, bulk_args_fun))
        task = asyncio.ensure_future(client.execute_many(stmt, args))
        await q.put(task)
    await q.put(None)


@argh.arg('--table', help='table name', required=True)
@argh.arg('--hosts', help='crate hosts', type=str)
@argh.arg('-n', '--num-records',
          help='number of records to insert',
          type=to_int,
          default=1e5)
@argh.arg('-b', '--bulk-size', type=to_int)
@argh.arg('-c', '--concurrency', type=to_int)
@argh.arg('--mapping-file',
          type=argparse.FileType('r'),
          help='JSON file with a column to fake provider mapping.')
@argh.wrap_errors([KeyboardInterrupt])
def insert_fake_data(hosts=None,
                     table=None,
                     num_records=1e5,
                     bulk_size=1000,
                     concurrency=25,
                     mapping_file=None):
    """Generate random data and insert it into a table.

    This will read the table schema and then find suitable random data providers.
    Which provider is choosen depends on the column name and data type.

    Example:

        A column named `name` will map to the `name` provider.
        A column named `x` of type int will map to `random_int` because there
        is no `x` provider.

    Available providers are listed here:
        http://fake-factory.readthedocs.io/en/latest/providers.html

        Additional providers:
        - auto_inc:
            Returns unique incrementing numbers.
            Automatically used for columns named "id" of type int or long
        - geo_point
            Returns [<lon>, <lat>]
            Automatically used for columns of type geo_point

    Args:
        hosts: <host>:[<port>] of the Crate node
        table: The table name into which the data should be inserted.
            Either fully qualified: `<schema>.<table>` or just `<table>`
        num_records: Number of records to insert.
            Usually a number but expressions like `1e4` work as well.
        bulk_size: The bulk size of the insert statements.
        concurrency: How many operations to run concurrently.
        mapping_file: A JSON file that defines a mapping from column name to
            fake-factory provider.
            The format is as follows:
            {
                "column_name": ["provider_with_args", ["arg1", "arg"]],
                "x": ["provider_with_args", ["arg1"]],
                "y": "provider_without_args"
            }
    """
    with connect(hosts) as conn:
        c = conn.cursor()
        schema, table_name = parse_table(table)
        columns = retrieve_columns(c, schema, table_name)
    if not columns:
        sys.exit('Could not find columns for table "{}"'.format(table))
    print('Found schema: ')
    print(json.dumps(columns, sort_keys=True, indent=4))
    mapping = None
    if mapping_file:
        mapping = json.load(mapping_file)

    bulk_size = min(num_records, bulk_size)
    num_inserts = int(num_records / bulk_size)

    generate_row = create_row_generator(columns, mapping)
    bulk_args_fun = partial(generate_bulk_args, generate_row, bulk_size)

    stmt = to_insert(table, columns)[0]
    print('Using insert statement: ')
    print(stmt)

    print('Will make {} requests with a bulk size of {}'.format(
        num_inserts, bulk_size))

    print('Generating fake data and executing inserts')
    q = asyncio.Queue(maxsize=concurrency)
    with clients.client(hosts, concurrency=concurrency) as client:
        loop.run_until_complete(asyncio.gather(
            _produce_data_and_insert(q, client, stmt, bulk_args_fun, num_inserts),
            consume(q, total=num_inserts)))


def main():
    argh.dispatch_command(insert_fake_data)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import argh
import itertools

from functools import partial
from time import time

from .cli import lines_from_stdin, to_int
from .clients import client
from .metrics import Stats
from . import aio


class Result:
    def __init__(self,
                 version_info,
                 statement,
                 meta,
                 started,
                 ended,
                 stats,
                 concurrency,
                 bulk_size=None,
                 output_fmt=None):
        self.version_info = version_info
        self.statement = str(statement)
        self.meta = meta
        # need ts in ms in crate
        self.started = int(started * 1000)
        self.ended = int(ended * 1000)
        self.runtime_stats = stats.get()
        self.concurrency = concurrency
        self.bulk_size = bulk_size

        # copy before str_func is assigned
        # because str_func shouldn't be part of the printed result
        self.d = self.__dict__.copy()
        output_fmt = output_fmt or 'full'
        if output_fmt == 'full':
            self.str_func = partial(self.as_json_string, self.d)
        elif output_fmt == 'short':
            self.str_func = partial(self.short_output, self.runtime_stats)
        else:
            raise ValueError('Invalid output format: {}'.format(output_fmt))

    @staticmethod
    def as_json_string(d):
        return json.dumps(d, sort_keys=True, indent=4)

    def as_dict(self):
        return self.d

    @staticmethod
    def format_stats(stats, output_fmt=None):
        output_fmt = output_fmt or 'full'
        if output_fmt == 'full':
            return Result.as_json_string(stats)
        else:
            return Result.short_output(stats)

    @staticmethod
    def short_output(stats):
        output = ('Runtime:\n'
                  '    mean: {mean:.3f} +/- {stdev:.3f}\n'
                  '    min:  {min:.3f}\n'
                  '    max:  {max:.3f}')
        values = dict(
            mean=stats['mean'],
            max=stats['max'],
            min=stats['min'],
            stdev=stats.get('stdev', 0.0)
        )
        if stats['n'] > 1:
            output += (
                '\n'
                'Percentile:\n'
                '    50:   {p50:.3f}\n'
                '    99.9: {p999:.3f}'
            )
            values.update(dict(
                p50=stats['percentile']['50'],
                p999=stats['percentile']['99_9']
            ))
        return output.format(**values)

    def __str__(self):
        return self.str_func()


class QueryRunner:
    def __init__(self,
                 stmt,
                 meta,
                 repeats,
                 hosts,
                 concurrency,
                 args=None,
                 bulk_args=None,
                 output_fmt=None):
        self.stmt = stmt
        self.meta = meta
        self.repeats = repeats
        self.concurrency = concurrency
        self.hosts = hosts
        self.client = client(hosts, concurrency=concurrency)
        self.bulk_args = bulk_args
        self.args = args
        self.output_fmt = output_fmt

    def warmup(self, num_warmup):
        statements = itertools.repeat((self.stmt,), num_warmup)
        aio.run_many(self.client.execute, statements, 0, num_items=num_warmup)

    def run(self):
        version_info = aio.run(self.client.get_server_version)
        started = time()
        if self.bulk_args:
            statements = itertools.repeat((self.stmt, self.bulk_args), self.repeats)
            f = self.client.execute_many
        else:
            statements = itertools.repeat((self.stmt, self.args), self.repeats)
            f = self.client.execute
        stats = Stats(min(self.repeats, 1000))
        measure = partial(aio.measure, stats, f)

        aio.run_many(
            measure, statements, self.concurrency, num_items=self.repeats)
        ended = time()

        return Result(
            statement=self.stmt,
            meta=self.meta,
            version_info=version_info,
            started=started,
            ended=ended,
            stats=stats,
            concurrency=self.concurrency,
            output_fmt=self.output_fmt
        )

    def __enter__(self):
        return self

    def __exit__(self, *ex):
        self.client.close()


@argh.arg('--hosts', help='crate hosts', type=str)
@argh.arg('-w', '--warmup', type=to_int)
@argh.arg('-r', '--repeat', type=to_int)
@argh.arg('-c', '--concurrency', type=to_int)
@argh.arg('-of', '--output-fmt', choices=['full', 'short'], default='full')
def timeit(hosts=None,
           stmt=None,
           warmup=30,
           repeat=30,
           concurrency=1,
           output_fmt=None):
    """ runs the given statement a number of times and returns the runtime stats
    """
    num_lines = 0
    for line in lines_from_stdin(stmt):
        with QueryRunner(line,
                         None,
                         repeat,
                         hosts=hosts,
                         concurrency=concurrency,
                         output_fmt=output_fmt) as runner:
            runner.warmup(warmup)
            result = runner.run()
        print(result)
        num_lines += 1
    if num_lines == 0:
        raise SystemExit(
            'No SQL statements provided. Use --stmt or provide statements via stdin')


def main():
    argh.dispatch_command(timeit)


if __name__ == '__main__':
    main()




import functools
import os
import asyncio
try:
    import uvloop
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    pass

from tqdm import tqdm
tqdm = functools.partial(
    tqdm,
    unit=' requests',
    smoothing=0.1,
    disable=os.environ.get('CR8_NO_TQDM') == 'True'
)


async def measure(stats, f, *args, **kws):
    duration = await f(*args, **kws)
    stats.measure(duration)
    return duration


async def qmap(q, corof, iterable):
    for i in iterable:
        task = asyncio.ensure_future(corof(*i))
        await q.put(task)
    await q.put(None)


async def consume(q, total=None):
    with tqdm(total=total) as t:
        while True:
            task = await q.get()
            if task is None:
                break
            await task
            t.update(1)


async def map(coro, iterable, total=None):
    for i in tqdm(iterable, total=total):
        await coro(*i)


def run(coro):
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(coro())


def run_many(coro, iterable, concurrency, num_items=None):
    loop = asyncio.get_event_loop()
    if concurrency == 1:
        return loop.run_until_complete(map(coro, iterable, total=num_items))
    q = asyncio.Queue(maxsize=concurrency)
    loop.run_until_complete(asyncio.gather(
        qmap(q, coro, iterable),
        consume(q, total=num_items)))



import argh
import os
import toml
from glob import glob
from .run_spec import run_spec
from .run_crate import CrateNode, get_crate


class Executor:
    def __init__(self, track_dir, result_hosts=None, crate_root=None, output_fmt=None, fail_fast=None):
        self.track_dir = track_dir
        self.result_hosts = result_hosts
        self.crate_root = crate_root
        self.output_fmt = output_fmt
        self.fail_fast = fail_fast

    def _expand_paths(self, paths):
        paths = (os.path.join(self.track_dir, path) for path in paths)
        paths = (os.path.abspath(path) for path in paths)
        return (p for path in paths for p in glob(path))

    def _run_specs(self, specs, benchmark_host):
        specs = self._expand_paths(specs)
        for spec in specs:
            print('### Running spec file: ', os.path.basename(spec))
            try:
                run_spec(
                    spec,
                    benchmark_host,
                    self.result_hosts,
                    output_fmt=self.output_fmt)
            except:
                if self.fail_fast:
                    raise
                else:
                    print('WARNING: Spec file failed due to the following exception:')
                    import traceback
                    traceback.print_exc()

    def execute(self, track):
        configurations = list(self._expand_paths(track['configurations']))
        versions = track['versions']
        for version in versions:
            print('# Version: ', version)
            for c, configuration in enumerate(configurations):
                print('## Starting Crate {0}, configuration: {1}'.format(
                    os.path.basename(version),
                    os.path.basename(configuration)
                ))
                configuration = toml.load(os.path.join(self.track_dir, configuration))
                crate_dir = get_crate(version, self.crate_root)
                with CrateNode(crate_dir=crate_dir,
                               env=configuration.get('env'),
                               settings=configuration.get('settings')) as node:
                    node.start()
                    self._run_specs(track['specs'], node.http_url)


@argh.arg('-r', '--result_hosts', type=str)
@argh.arg('-of', '--output-fmt', choices=['full', 'short'], default='full')
@argh.arg('--failfast', action='store_true')
@argh.wrap_errors([KeyboardInterrupt])
def run_track(track, result_hosts=None, crate_root=None, output_fmt=None, failfast=False):
    """Execute a track file"""
    executor = Executor(
        track_dir=os.path.dirname(track),
        result_hosts=result_hosts,
        crate_root=crate_root,
        output_fmt=output_fmt,
        fail_fast=failfast
    )
    executor.execute(toml.load(track))


def main():
    argh.dispatch_command(run_track)


if __name__ == "__main__":
    main()



from unittest import TestCase
from cr8.timeit import Result
from cr8.metrics import Stats


class ResultTest(TestCase):

    def test_short_result_output_with_only_1_measurement(self):
        stats = Stats(1)
        stats.measure(23.4)
        result = Result(
            version_info={},
            meta=None,
            statement='select name from sys.cluster',
            started=10,
            ended=20,
            stats=stats,
            concurrency=1,
            output_fmt='short')
        self.assertEqual(
            str(result),
            ('Runtime:\n'
             '    mean: 23.400 +/- 0.000\n'
             '    min:  23.400\n'
             '    max:  23.400')
        )

    def test_short_result_output_with_more_measurements(self):
        stats = Stats(4)
        stats.measure(23.4)
        stats.measure(48.7)
        stats.measure(32.5)
        stats.measure(15.9)
        result = Result(
            version_info={},
            meta=None,
            statement='select name from sys.cluster',
            started=10,
            ended=20,
            stats=stats,
            concurrency=1,
            output_fmt='short')
        self.assertEqual(
            str(result),
            ('Runtime:\n'
             '    mean: 30.125 +/- 14.121\n'
             '    min:  15.900\n'
             '    max:  48.700\n'
             'Percentile:\n'
             '    50:   23.400\n'
             '    99.9: 48.700')
        )
